class RegionException(Exception):
    pass


class FrameException(Exception):
    pass


class ArrayException(Exception):
    pass


class GridException(Exception):
    pass


class VectorFieldException(Exception):
    pass


class KernelException(Exception):
    pass


class ConvolverException(Exception):
    pass


class MaskException(Exception):
    pass


class MappingException(Exception):
    pass


class DatasetException(Exception):
    pass


class PixelizationException(Exception):
    pass


class InversionException(Exception):
    pass


class PlottingException(Exception):
    pass
